@@ -17,3 +17,8 @@ versions = get_versions()
 __version__ = versions['version']
 __git_revision__ = versions['full-revisionid']
 del get_versions, versions
+
+# code that allows access to the data directory
+_ROOT = os.path.abspath(os.path.dirname(__file__))
+def get_directory(path='../data'):
+    return os.path.join(_ROOT, path)
